//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef PHOBJ_HPP
#define PHOBJ_HPP
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
//------------------------------------------------------------------------------
#include <utility>
#include <vector>
//------------------------------------------------------------------------------
#include "Object.hpp"
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
class LinePtPt;
//------------------------------------------------------------------------------
class LinePtVect
{
public:
    LinePtVect(const Vector2D &cPt, const Vector2D &cVect);
    LinePtVect(const LinePtPt &line);
    float GetSize() const;
    float GetSqrSum() const;

public:
    Vector2D pt;
    Vector2D vect;
};
//------------------------------------------------------------------------------
class LinePtPt
{
public:
    LinePtPt(const Vector2D &cPt1, const Vector2D &cPt2);
    LinePtPt(const LinePtVect &line);
    float GetSize() const;
    float GetSqrSum() const;

public:
    Vector2D pt1;
    Vector2D pt2;
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A base class containing common functions to all physic objects. */
/// LUA_HEADER_COPY_OFF ///
class PhysicObject : public Object
{
public:
    PhysicObject(Group *parent, int phGroupId, const std::string &name);
    //PhysicObject(const PhysicObject& src);
    virtual ~PhysicObject();

    /// LUA_EXPORT_ON(PhysicObject, Object, ProtectedPtr<Object>) ///
    /** Return true if the point \\a is inside the object, false otherwise. */
    virtual bool IsPointInside(const Vector2D &point) = 0;

    // Lua
    /** Acces the event function name (\\a fnName should be the name of an existing lua function). */
    //LUA// luaProperty(std::string, eventFunction, GetEventFunction, SetEventFunction);
    /** Set the event function name (\\a fnName should be the name of an existing lua function).\\n
      * Function prototype should be : function fnName(thisObject : PhysicObject, theOtherObject : PhysicObject, intersectionPosition : Vector2D, penetrationVector : Vector2D)\\n
      * Where :
      * <ul>
      *     <li><strong>thisObject</strong> is this physic object</li>
      *     <li><strong>theOtherObject</strong> is physic object which intersect this one</li>
      *     <li><strong>intersectionPosition</strong> the global position of the intersection</li>
      *     <li><strong>penetrationVector</strong> how much you need to move <strong>thisObject</strong> to remove it from <strong>theOtherObject</strong></li>
      * </ul>
      */
    void SetEventFunction(const luabind::object &luaFunction);
    /** Get the event function name. */
    luabind::object GetEventFunction() const;

    /** Get the physic group associated to the object */
    int GetPhysicGroup() const;

    /** Return true if the object collide the given object */
    bool IsIntersect(const PhysicObject &obj) const;

    /** Return the list of objects that collide the given physic group */
    luabind::object IsIntersectWithGroup(int phGroup);
    /// LUA_EXPORT_OFF ///


    // No Lua
    // Lua
    //static void RegisterClassLua(lua_State *myLuaState);

    // Intersection
    virtual bool Intersect(const PhysicObject &obj, Vector2D &pos, Vector2D &penetrationVector) const;

    // Event
    void OnIntersect(PhysicObject *obj1, PhysicObject *obj2, const Vector2D &pos, const Vector2D &penetrationVector);

    // Parent
    void ChangeParent(Group *parent);

    // Show
    virtual void RenderShape(sf::RenderTarget &App) const = 0;

    // Maximum interaction radius
    virtual float GetMaxInteractionRadius() const = 0;

    // Some object need to be noticed when the parent screen is changed
    virtual void OnChangeScreen(Screen *backScreen, Screen *newScreen);

    // Register a clone
    virtual void RegisterClone(Group *newParent);
protected:

    int m_group;
    luabind::object m_luaFunction;

    bool IsLineInter(const LinePtVect &l1, const LinePtVect &l2, Vector2D &pos) const;
    Vector2D GetIntersectionPoint(const Vector2D &pos1, const Vector2D &direction1, const Vector2D &pos2, const Vector2D &direction2) const;
    //bool IsLineCircleInter(const LinePtVect &l1, const Vector2D &center, float radius, Vector2D &pos) const;

protected:
    struct minMaxStr
    {
        float minVal, maxVal;
    };
    //------------------------------------------------------------------------------
    struct AxisVectorData
    {
        minMaxStr minMax1, minMax2;
    };
    //------------------------------------------------------------------------------
    struct vector2Dcompar
    {
        bool operator() (const Vector2D& lv, const Vector2D& rv) const
        {
            if(lv.x < rv.x)
            {
                return true;
            }
            else if(lv.x == rv.x)
            {
                return lv.y < rv.y;
            }
            else
            {
                return false;
            }
        }
    };
    //------------------------------------------------------------------------------
    typedef std::map<Vector2D, AxisVectorData, vector2Dcompar> axisListType;
    //------------------------------------------------------------------------------
    /** Projection d'un point sur un axe */
    void projectPointToAxis(const Vector2D &axis, minMaxStr &minMax, const Vector2D &point) const;

    /** Projection d'une forme en lignes sur un axe */
    minMaxStr ProjectLineShapeToAxis(const std::vector<Vector2D> &ptList, const Vector2D& axis) const;

    /** Projection d'un cercle sur un axe */
    minMaxStr ProjectCircleToAxis(const Vector2D &center, float radius, const Vector2D& axis) const;

    /** Get the collision point and the penetration vector (To get obj2 out of obj1) */
    void GetCollisionInfoFrom(const PhysicObject &obj1, const PhysicObject &obj2, const axisListType &axisList, Vector2D &point, Vector2D &penetrationVector) const;

    /** Create axis list from a point list and add it to the given list */
    void GetAxisFromLineShape(const std::vector<Vector2D> &ptList, axisListType &axisList) const;

    /** Create axis list from a cirlce and the point list of the other object and add it to the given list */
    void GetAxisFromDisk(const Vector2D &center, const std::vector<Vector2D> &objPtList, axisListType &axisList) const;

    /** Return a vector that alwas have x > 0 and (y > 0 if x == 0) */
    Vector2D GetSameDirectionVector(Vector2D vect) const;
};
//------------------------------------------------------------------------------
class PhysicObjectLine : public PhysicObject
{
public:
    // No Lua
    PhysicObjectLine(Group *parent, int phGroupId, const std::string &name) : PhysicObject(parent, phGroupId, name) {}
    virtual bool IsPointInside(const Vector2D &point);

    virtual std::vector<Vector2D> GetAllPoints() const = 0;
    virtual void RenderShape(sf::RenderTarget &App) const;

    // Maximum interaction radius
    virtual float GetMaxInteractionRadius() const {return m_cachedMaxInterRadius;}

protected:
    float m_cachedMaxInterRadius;
    void RefreshMaxInterRadius();
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A structure containing informations about a collision. */
/// LUA_HEADER_COPY_OFF ///
struct CollisionInfo
{
public:
    // Constructor
    CollisionInfo(PhysicObject *pPhObj, const Vector2D &pPoint, const Vector2D &pPenetrationVector) : phObj(pPhObj), point(pPoint), penetrationVector(penetrationVector) {}

    /// LUA_EXPORT_ON(CollisionInfo) ///
    /** Get the PhysicObject with collide with the other object */
    //LUA// luaReadwrite(PhysicObject*, phObj);

    /** Get the collision point */
    //LUA// luaReadwrite(Vector2D, point);

    /** Get the penetration vector (to get the tested object out of phObj */
    //LUA// luaReadwrite(Vector2D, penetrationVector);
    /// LUA_EXPORT_OFF ///

    PhysicObject *phObj;
    Vector2D point;
    Vector2D penetrationVector;
};
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // PHOBJ_HPP
