//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Text.hpp"

#include <iostream>
#include <string>
#include <algorithm>

#include "Utils.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
Text::Text(Group *parent, const std::string &name, const std::string &str) : DrawableSF(parent, name), sf::String(""), m_isChanged(true), m_canCutWords(false), m_maxWidth(-1), m_originalText(str)
{
    ArrangeTextForWidth();
}
//------------------------------------------------------------------------------
void Text::ProceedMvt(float time)
{
    DrawableSF::ProceedMvt(time);

    if(m_isChanged)
    {
        ArrangeTextForWidth();
    }
}
//------------------------------------------------------------------------------
Vector2D Text::GetTextSize()
{
    if(m_isChanged)
    {
        ArrangeTextForWidth();
    }

    sf::Rect<float> rect = GetRect();
    return Vector2D(rect.GetWidth(), rect.GetHeight());
}
//------------------------------------------------------------------------------
void Text::ArrangeTextForWidth()
{
    m_isChanged = false; // We need to set m_isChanged to false at the begining so that ArrangeTextForWidth() won't ba called in GetTextSize()

    if(m_maxWidth <= 0)
    {
        sf::String::SetText(StrToSfUtf8(m_originalText));
        return;
    }

    string newText, currentLine;

    if(m_canCutWords)
    {
        for(string::const_iterator it = m_originalText.begin(); it != m_originalText.end(); it++)
        {
            currentLine += *it;
            sf::String::SetText(StrToSfUtf8(currentLine)); // We use the current object to mesure the line width.
            if(GetTextSize().x > m_maxWidth && currentLine.size() > 1) // If we are longer than the limit AND the current line isn't empty (prevent endless loop)
            {
                newText.append(currentLine.begin(), currentLine.end() - 1); // We skip the last char (witch makes the line longer than the limit)
                newText += '\n';
                currentLine.clear();
                currentLine += *it; // We add the skipped char
            }
        }
        newText += currentLine; // We add the unfinished line
    }
    else
    {
        string::const_iterator spacePos;
        for(string::const_iterator it = m_originalText.begin(); it != m_originalText.end(); it = spacePos)
        {
            spacePos = std::find(it + 1, string::const_iterator(m_originalText.end()), ' ');
            size_t wordSize = spacePos - it;

            currentLine.append(it, spacePos);
            sf::String::SetText(StrToSfUtf8(currentLine)); // We use the current object to mesure the line width.
            if(GetTextSize().x > m_maxWidth && currentLine.size() > wordSize) // If we are longer than the limit AND the current line isn't empty (prevent endless loop)
            {
                newText.append(currentLine.begin(), currentLine.end() - wordSize); // We skip the last word (witch makes the line longer than the limit)
                newText += '\n';
                currentLine.clear();
                if(*it == ' ')
                {
                    it++; // We wkip the space (if any) at the begining of the new line
                }
                currentLine.append(it, spacePos); // We add the skipped word
            }
        }
        newText += currentLine; // We add the unfinished line
    }

    sf::String::SetText(StrToSfUtf8(newText));
}

//------------------------------------------------------------------------------
/** Set if the Text should be rendered bold */
void Text::SetBold(bool isBold)
{
	if(isBold)
	{
	    SetStyle(GetStyle() | sf::String::Bold);
	}
	else
	{
	    SetStyle(GetStyle() & ~(sf::String::Bold));
	}

	m_isChanged = true;
}
//------------------------------------------------------------------------------
/** Tells if the Text should be rendered bold */
bool Text::GetBold() const
{
    return GetStyle() & sf::String::Bold;
}
//------------------------------------------------------------------------------

/** Set if the Text should be rendered italic */
void Text::SetItalic(bool isItalic)
{
	if(isItalic)
	{
	    SetStyle(GetStyle() | sf::String::Italic);
	}
	else
	{
	    SetStyle(GetStyle() & ~(sf::String::Italic));
	}

	m_isChanged = true;
}
//------------------------------------------------------------------------------
/** Tells if the Text should be rendered italic */
bool Text::GetItalic() const
{
    return GetStyle() & sf::String::Italic;
}
//------------------------------------------------------------------------------

/** Set if the Text should be rendered underligned */
void Text::SetUnderligned(bool isUnderligned)
{
	if(isUnderligned)
	{
	    SetStyle(GetStyle() | sf::String::Underlined);
	}
	else
	{
	    SetStyle(GetStyle() & ~(sf::String::Underlined));
	}

	m_isChanged = true;
}
//------------------------------------------------------------------------------
/** Tells if the Text should be rendered underligned */
bool Text::GetUnderligned() const
{
    return GetStyle() & sf::String::Underlined;
}
//------------------------------------------------------------------------------
void Text::Render(sf::RenderTarget &App)
{
    // Check for invalidated pointers
    if(!AreResourcesOk() && !GetResource().IsNull())
    {
        MY_THROW("Text::Render(): the font used by the Text was deleted ! Object name: '" << GetName() << "'.")
    }

    // Render
    RenderSf(App, this);
}
//------------------------------------------------------------------------------
void Text::SetFont(Font &charFont)
{
    ProtectedPtr<ProtectedClassBase> ptr(&charFont);

    if(!ptr.IsValid())
    {
        MY_THROW("Sprite::SetImage(): the texture was deleted ! Object name: '" << GetName() << "'.");
    }

    // Ressources
    SetResource(ptr);

    // Font
    sf::String::SetFont(charFont);

    // Invalidate
    InvalidateText();
}
//------------------------------------------------------------------------------
ProtectedPtr<Font> Text::GetFont()
{
    if(!AreResourcesOk())
    {
        MY_THROW("Text::GetFont(): You are trying to get a null or deleted font ! Object name: '" << GetName() << "'.")
    }

    return ProtectedPtr<Font>(static_cast<Font*>(GetResource().GetPtr()));
}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
